﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace MarkusGrothoff.Media {

    /// <summary>
    /// Stellen Player dar, die Multimedia-Dateien wiedergeben.
    /// </summary>
    public class MediaPlayer : IDisposable {

        /// <summary>
        /// Datei wurde abgespielt.
        /// </summary>
        public event EventHandler Played = null;

        private static readonly ICollection<String> aliasInUse = new List<String>();

        private readonly IntPtr displayHandle;
        private String recentlyUsedCommand = String.Empty;

        /// <summary>
        /// Ruft den Dateinamen inkl. Pfad der geöffneten Datei ab.
        /// Ist keine Datei geöffnet, wird String.Empty zurückgegeben.
        /// </summary>
        public String FileName { get; private set; }

        /// <summary>
        /// Ruft ab, ob das Gerät geöffnet ist.
        /// </summary>
        protected bool IsOpen { get; private set; }

        /// <summary>
        /// Ruft den Alias für die Datei ab.
        /// </summary>
        protected String Alias { get; private set; }

        /// <summary>
        /// Ruft das Control ab, welches die Fensternachrichten empfängt.
        /// </summary>
        private MciCallbackControl CallbackControl { get; set; }

        /// <summary>
        /// Ruft das Handle des Callback-Controls ab.
        /// </summary>
        protected IntPtr CallbackHandle { get { return CallbackControl.Handle; } }

        /// <summary>
        /// Erstellt einen neuen Player.
        /// </summary>
        /// <param name="handle">Control-Handle für die Darstellung von Videos.</param>
        protected MediaPlayer(IntPtr displayHandle) {
            this.displayHandle = displayHandle;
            CallbackControl = new MciCallbackControl();
            CallbackControl.NotifyMessageReceived += new MciCallbackControl.NotifyMessageReceivedHandler(CallbackControl_NotifyMessageReceived);
            FileName = String.Empty;
            IsOpen = false;
            Alias = String.Empty;
        }

        /// <summary>
        /// Erstellt einen neuen Player.
        /// </summary>
        public MediaPlayer() : this(IntPtr.Zero) { }

        /// <summary>
        /// Öffnet eine Datei.
        /// </summary>
        /// <param name="fileName">Pfad- und Dateiname.</param>
        /// <exception cref="InvalidOperationException">Es ist bereits eine Datei von diesem Player geöffnet.</exception>
        /// <exception cref="FileNotFoundException">Datei wurde nicht gefunden.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public void Open(String fileName) {
            if (IsOpen) {
                throw new InvalidOperationException("One file is already opened by this player.");
            } else if (!File.Exists(fileName)) {
                throw new FileNotFoundException("File couldn't be found.", fileName);
            } else {

                // Datei öffnen.
                String shortFileName = String.Format(@"""{0}""", fileName);
                String alias = GetAlias(shortFileName);
                String command = "open " + shortFileName + " alias " + alias;
                int result = MciApi.SendString(command, null, 0, CallbackHandle);

                if (result != 0) {
                    // Datei konnte nicht geöffnet werden.
                    throw CreateMciException(result);
                } else {

                    // Zeitformat auf Stunden, Minuten, Sekunden setzen.
                    result = MciApi.SendString("set " + alias + " time format milliseconds",
                        null, 0, CallbackHandle);

                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        // Falls es sich um eine Video-Datei handelt.
                        if (displayHandle != IntPtr.Zero) MciApi.SendString("window " + alias + " handle " + displayHandle.ToInt32(),
                              null, 0, CallbackHandle);
                        FileName = fileName;
                        this.Alias = alias;
                        IsOpen = true;
                        MediaPlayer.aliasInUse.Add(this.Alias);
                    }
                }
            }
        }

        /// <summary>
        /// Schließt die Datei, falls eine geöffnet ist.
        /// </summary>
        public void Close() {
            if (IsOpen) {
                MciApi.SendString("close " + Alias, null, 0, CallbackHandle);
                MediaPlayer.aliasInUse.Remove(Alias);
                Alias = String.Empty;
                IsOpen = false;
                FileName = string.Empty;
            }
        }

        /// <summary>
        /// Ruft den Status des Players ab.
        /// </summary>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public MediaStatus Status {
            get {
                if (!IsOpen) {
                    return MediaStatus.Closed;
                } else {
                    StringBuilder returnString = new StringBuilder(1024);
                    int result = MciApi.SendString("status " + Alias + " mode", returnString,
                        returnString.Capacity, CallbackHandle);
                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        switch (returnString.ToString()) {
                            case "stopped":
                                return MediaStatus.Stopped;
                            case "playing":
                                return MediaStatus.Playing;
                            case "paused":
                                return MediaStatus.Paused;
                            default:
                                return MediaStatus.Stopped;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// Startet die Wiedergabe an der gewünschten Position.
        /// </summary>
        /// <param name="from">Startposition.</param>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Die Position ist ungültig, weil sie
        /// kleiner 0 oder größer der Gesamtspiellänge ist.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public void Play(TimeSpan from) {
            if (!IsOpen) {
                throw new InvalidOperationException("No file is opened.");
            } else if (from < TimeSpan.Zero || from > Length) {
                throw new ArgumentOutOfRangeException("from", from, "Invalid position.");
            } else {
                int result = MciApi.SendString("play " + Alias + " from " + from.TotalMilliseconds + " notify",
                    null, 0, CallbackHandle);
                if (result != 0) {
                    throw CreateMciException(result);
                } else {
                    recentlyUsedCommand = "play";
                }
            }
        }

        /// <summary>
        /// Startet die Wiedergabe vom Anfang der Datei.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public void Play() { Play(TimeSpan.Zero); }

        /// <summary>
        /// Stoppt die Wiedergabe.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public void Stop() {
            if (!IsOpen) {
                throw new InvalidOperationException("No file is opened.");
            } else {
                int result = MciApi.SendString("stop " + Alias, null, 0, CallbackHandle);
                if (result != 0) throw CreateMciException(result);
            }
        }

        /// <summary>
        /// Pausiert die Wiedergabe.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public void Pause() {
            if (!IsOpen) {
                throw new InvalidOperationException("No file is opened.");
            } else {
                int result = MciApi.SendString("pause " + Alias, null, 0, CallbackHandle);
                if (result != 0) throw CreateMciException(result);
            }
        }

        /// <summary>
        /// Kehrt zur Wiedergabe zurück.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public void Resume() {
            if (!IsOpen) {
                throw new InvalidOperationException("No file is opened.");
            } else {
                int result = MciApi.SendString("resume " + Alias, null, 0, CallbackHandle);
                if (result != 0) throw CreateMciException(result);
            }
        }

        /// <summary>
        /// Ruft ab, ob die Audiowiedergabe stummgeschaltet ist, oder legt es fest.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public bool IsMuted {
            get {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else {
                    StringBuilder returnString = new StringBuilder(1024);
                    int result = MciApi.SendString("status " + Alias + " audio", returnString,
                        returnString.Capacity, CallbackHandle);
                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        return returnString.ToString() == "off";
                    }
                }
            }

            set {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else {
                    String flag = value ? "off" : "on";
                    int result = MciApi.SendString("set " + Alias + " audio all " + flag, null, 0, CallbackHandle);
                    if (result != 0) throw CreateMciException(result);
                }
            }
        }

        /// <summary>
        /// Ruft die Lautstärke beider Kanäle ab oder legt sie fest.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Lautstärke muss zwischen 0.0 und 1.0 liegen.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public Double Volume {
            get {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else {
                    StringBuilder returnString = new StringBuilder(1024);
                    int result = MciApi.SendString("status " + Alias + " volume", returnString,
                        returnString.Capacity, CallbackHandle);
                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        return double.Parse(returnString.ToString()) * 0.001;
                    }
                }
            }

            set {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else if (value < 0.0 || value > 1.0) {
                    throw new ArgumentOutOfRangeException("value", value, "Invalid volume.");
                } else {
                    int volume = (int)(value * 1000.0);
                    int result = MciApi.SendString("setaudio " + Alias + " volume to " + volume.ToString(),
                        null, 0, CallbackHandle);
                    if (result != 0) throw CreateMciException(result);
                }
            }
        }

        /// <summary>
        /// Ruft die Lautstärke des linken Kanals ab oder legt sie fest.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Lautstärke muss zwischen 0.0 und 1.0 liegen.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public Double LeftVolume {
            get {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else {
                    StringBuilder returnString = new StringBuilder(1024);
                    int result = MciApi.SendString("status " + Alias + " left volume", returnString,
                        returnString.Capacity, CallbackHandle);
                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        return double.Parse(returnString.ToString()) * 0.001;
                    }
                }
            }

            set {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else if (value < 0.0 || value > 1.0) {
                    throw new ArgumentOutOfRangeException("value", value, "Invalid volume.");
                } else {
                    int volume = (int)(value * 1000.0);
                    int result = MciApi.SendString("setaudio " + Alias + " left volume to " + volume.ToString(),
                        null, 0, CallbackHandle);
                    if (result != 0) throw CreateMciException(result);
                }
            }
        }

        /// <summary>
        /// Ruft die Lautstärke des rechten Kanals ab oder legt sie fest.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Lautstärke muss zwischen 0.0 und 1.0 liegen.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public Double RightVolume {
            get {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else {
                    StringBuilder returnString = new StringBuilder(1024);
                    int result = MciApi.SendString("status " + Alias + " right volume", returnString,
                        returnString.Capacity, CallbackHandle);
                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        return double.Parse(returnString.ToString()) * 0.001;
                    }
                }
            }

            set {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else if (value < 0.0 || value > 1.0) {
                    throw new ArgumentOutOfRangeException("value", value, "Invalid volume.");
                } else {
                    int volume = (int)(value * 1000.0);
                    int result = MciApi.SendString("setaudio " + Alias + " right volume to " + volume.ToString(),
                        null, 0, CallbackHandle);
                    if (result != 0) throw CreateMciException(result);
                }
            }
        }

        /// <summary>
        /// Ruft die Position in der Datei ab oder legt sie fest.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="ArgumentOutOfRangeException">Die Position ist ungültig, weil sie
        /// kleiner 0 oder größer der Gesamtspiellänge ist.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public TimeSpan Position {
            get {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else {
                    StringBuilder returnString = new StringBuilder(1024);
                    int result = MciApi.SendString("status " + Alias + " position", returnString,
                        returnString.Capacity, CallbackHandle);
                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        return TimeSpan.FromMilliseconds(double.Parse(returnString.ToString()));
                    }
                }
            }

            set {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else if (value < TimeSpan.Zero || value > Length) {
                    throw new ArgumentOutOfRangeException("value", value, "Invalid position.");
                } else {
                    if (Status != MediaStatus.Playing) {
                        int result = MciApi.SendString("seek " + Alias + " to " + value.TotalMilliseconds,
                            null, 0, CallbackHandle);
                        if (result != 0) throw CreateMciException(result);
                    } else {
                        Play(value);
                    }
                }
            }
        }

        /// <summary>
        /// Ruft die Gesamtspiellänge der Datei ab.
        /// </summary>
        /// <exception cref="InvalidOperationException">Keine Datei geöffnet.</exception>
        /// <exception cref="MediaException">Der Befehl konnte nicht ausgeführt werden.</exception>
        public TimeSpan Length {
            get {
                if (!IsOpen) {
                    throw new InvalidOperationException("No file is opened.");
                } else {
                    StringBuilder returnString = new StringBuilder(1024);
                    int result = MciApi.SendString("status " + Alias + " length", returnString,
                        returnString.Capacity, CallbackHandle);
                    if (result != 0) {
                        throw CreateMciException(result);
                    } else {
                        return TimeSpan.FromMilliseconds(double.Parse(returnString.ToString()));
                    }
                }
            }
        }

        /// <summary>
        /// Erstellt eine Mci-Ausnahme, die aus dem HResult-Wert generiert wird.
        /// </summary>
        /// <param name="hResult">Rückgabewert der Api-Funktion.</param>
        /// <returns>Mci-Ausnahme.</returns>
        protected MediaException CreateMciException(int hResult) {
            StringBuilder errorText = new StringBuilder(1024);
            MciApi.GetErrorString(hResult, errorText, errorText.Capacity);
            return new MediaException(errorText.ToString());
        }

        /// <summary>
        /// Holt einen eindeutigen Alias für die Datei.
        /// </summary>
        /// <param name="fileName">Pfad- und Dateiname.</param>
        /// <returns>Eindeutiger Alias.</returns>
        private String GetAlias(String fileName) {
            String alias = fileName.GetHashCode().ToString();
            while (MediaPlayer.aliasInUse.Contains(alias)) {
                alias = alias.GetHashCode().ToString();
            }
            return alias;
        }

        private void CallbackControl_NotifyMessageReceived(object sender, NotifyMessageReceivedEventArgs e) {
            if (e.Message == MciApi.NotifyMessage.Successful) {
                switch (recentlyUsedCommand) {
                    case "play":
                        OnPlayed(new EventArgs());
                        break;
                }
            }
            recentlyUsedCommand = String.Empty;
        }

        /// <summary>
        /// Löst das Played-Event aus.
        /// </summary>
        protected virtual void OnPlayed(EventArgs e) {
            if (Played != null) {
                EventHandler played = (EventHandler)Played.Clone();
                if (played != null) played(this, e);
            }
        }

        #region IDisposable Member

        /// <summary>
        /// Gibt alle vom Objekt benutzten Resourcen frei.
        /// </summary>
        public void Dispose() {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ~MediaPlayer() {
            Dispose(false);
        }

        protected virtual void Dispose(bool disposing) {
            if (disposing) {
            }
            Close();
        }

        #endregion
    }
}
